#ifndef PM_POWERMAN_H
#define PM_POWERMAN_H

#define DAEMON_NAME         "powermand"
#define DFLT_PORT           "10101"
#define DFLT_HOSTNAME       "127.0.0.1"

#endif /* PM_POWERMAN_H */

/*
 * vi:tabstop=4 shiftwidth=4 expandtab
 */
